#pragma once

class CClass;

class DynamicMethodsPrototype
{
	// I\bh
	Methods dynamicMethods;

	// XMLVACYp
private:
	friend class boost::serialization::access;
	template<class Archive> void serialize(Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP( dynamicMethods );
	}

public:
	DynamicMethodsPrototype(){}
	DynamicMethodsPrototype( const DynamicMethodsPrototype &dynamicMethodsPrototype )
		: dynamicMethods( dynamicMethodsPrototype.dynamicMethods )
	{
	}
	~DynamicMethodsPrototype(){}

	const Methods &GetDynamicMethods() const
	{
		return dynamicMethods;
	}
	Methods &GetDynamicMethods()
	{
		return dynamicMethods;
	}

	void AddDynamicMethods( CMethod *pMethod )
	{
		dynamicMethods.push_back( pMethod );
	}
};

class Interface : public DynamicMethodsPrototype
{
	const CClass *pInterfaceClass;
	mutable LONG_PTR vtblOffset;

	// ^p[^ip[^j
	Types actualTypeParameters;

	// XMLVACYp
private:
	friend class boost::serialization::access;
	template<class Archive> void serialize(Archive& ar, const unsigned int version)
	{
		trace_for_serialize( "serializing - Interface" );

		ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP( DynamicMethodsPrototype );
		ar & boost::serialization::make_nvp("pInterfaceClass", const_cast<CClass *&>(pInterfaceClass) );
		ar & BOOST_SERIALIZATION_NVP( vtblOffset );
		ar & BOOST_SERIALIZATION_NVP( actualTypeParameters );
	}

public:
	Interface( const CClass *pInterfaceClass, const Types &actualTypeParameters );
	Interface( const Interface &objInterface )
		: DynamicMethodsPrototype( objInterface )
		, pInterfaceClass( objInterface.pInterfaceClass )
		, vtblOffset( objInterface.vtblOffset )
	{
	}
	Interface()
	{
	}

	const CClass &GetClass() const{
		return *pInterfaceClass;
	}
	LONG_PTR GetVtblOffset() const
	{
		return vtblOffset;
	}
	void SetVtblOffset( LONG_PTR vtblOffset ) const
	{
		this->vtblOffset = vtblOffset;
	}

	const Types &GetActualTypeParameters() const
	{
		return actualTypeParameters;
	}

	std::string GetFullNameWithActualGenericTypeParameters() const;
};
typedef std::vector<Interface *> Interfaces;
